# -*- coding: utf-8 -*-
"""
Created on Wed May 19 14:25:56 2021

@author: ormondt
"""
import pandas as pd
import cht_tide.constituent as cons
from cht_tide.tide import Tide


def predict(data, times, format="np"):
    all_constituents = [c for c in cons.noaa if c != cons._Z0]
    constituents = []
    amplitudes   = []
    phases       = []
    for name in data.index.to_list():
        okay = False
        noaa_name = name
        if name == "MM":
            noaa_name = "Mm"
        if name == "MF":
            noaa_name = "Mf"
        if name == "SA":
            noaa_name = "Sa"
        if name == "SSA":
            noaa_name = "Ssa"
        if name == "MU2":
            noaa_name = "mu2"
        if name == "NU2":
            noaa_name = "nu2"
        for cnst in all_constituents:
            if cnst.name == noaa_name:
                constituents.append(cnst)
                # Check if amplitude is a column in the data
                if "amplitude" in data.columns:
                    amplitudes.append(data.loc[name, "amplitude"])
                else:
                    # Assume it is the first non-index column
                    amplitudes.append(data.loc[name, 1])
                if "phase" in data.columns:
                    phases.append(data.loc[name, "phase"])
                else:
                    # Assume it is the second non-index column
                    phases.append(data.loc[name, 2])    
                okay = True                
                continue
        if not okay:    
            print(f"Constituent {name} not found in list of NOAA constituents ! Skipping ...")    

    td = Tide(
        constituents=constituents,
        amplitudes=amplitudes,
        phases=phases,
    )
    v = td.at(times)

    if format == "dataframe" or format == "df":
        # Convert numpy array v to dataframe where index is time
        v = pd.DataFrame(v, index=times)

    return v



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 10:58:08 2021

@author: Maarten van Ormondt
"""

import os
import yaml
import toml
import boto3
from botocore import UNSIGNED
from botocore.client import Config

from .fes2014 import TideModelFes2014

class TideModelDatabase:
    """
    The main Tide Model Database class
    
    :param pth: Path name where bathymetry tiles will be cached.
    :type pth: string            
    """
    
    def __init__(self, path=None,
                 s3_bucket=None,
                 s3_key=None,
                 s3_region=None):
        self.path    = path
        self.dataset = []
        self.s3_client = None
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.s3_region = s3_region
        self.read()
    
    def read(self):
        """
        Reads meta-data of all datasets in the database. 
        """
        if self.path is None:
            print("Path to tide model database not set !")
            return
        
        # Check if the path exists. If not, create it.
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        # Read in database
        tml_file = os.path.join(self.path, "tide_models.tml")
        if not os.path.exists(tml_file):
            print("Warning! Tide model database file not found: " + tml_file)
            return

        datasets = toml.load(tml_file)

        for d in datasets["dataset"]:

            name = d["name"]

            if "path" in d:
                path = d["path"]
            else:
                path = os.path.join(self.path, name)

            # Read the meta data for this dataset
            fname = os.path.join(path, "metadata.tml")

            if os.path.exists(fname):
                metadata = toml.load(fname)
                dataset_format = metadata["format"]
            else:
                print("Could not find metadata file for dataset " + name + " ! Skipping dataset.")
                continue

            if dataset_format.lower() == "fes2014":
                model = TideModelFes2014(name, path)
            elif dataset_format.lower() == "tpxo_old":
                pass
            
            self.dataset.append(model)

    def check_online_database(self):
        if self.s3_client is None:
            self.s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        if self.s3_bucket is None:
            return
        # First download a copy of bathymetry.tml and call it bathymetry_s3.tml
        key = f"{self.s3_key}/tide_models.tml"
        filename = os.path.join(self.path, "tide_models_s3.tml")
        print("Updating tide models database ...")
        try:
            self.s3_client.download_file(Bucket=self.s3_bucket,     # assign bucket name
                                         Key=key,           # key is the file name
                                         Filename=filename) # storage file path
        except:
            # Download failed
            print(f"Failed to download {key} from {self.s3_bucket}. Database will not be updated.")
            return

        # Read bathymetry_s3.tml
        short_name_list, long_name_list = self.dataset_names()
        datasets_s3 = toml.load(filename)
        tide_models_added = False
        added_names = []
        # Loop through s3 datasets, and check whether they exist in the local database.
        # If so, check if the metadata also exists. If not, make local folder and download the metadata.
        # Additionally, check if available_tiles.nc in s3 and not in local database, download it.
        for d in datasets_s3["dataset"]:
            # Get list of existing datasets
            s3_name = d["name"]
            if s3_name not in short_name_list:
                # Dataset not in local database
                print(f"Adding tide model {s3_name} to local database ...")
                # Create folder and download metadata
                path = os.path.join(self.path, s3_name)
                os.makedirs(path, exist_ok=True)
                key = f"{self.s3_key}/{s3_name}/metadata.tml"
                filename = os.path.join(path, "metadata.tml")
                # Download metadata
                try:
                    self.s3_client.download_file(Bucket=self.s3_bucket, # assign bucket name
                                                Key=key,               # key is the file name
                                                Filename=filename)     # storage file path
                except Exception as e:
                    print(e)
                    print(f"Failed to download {key}. Skipping tide model.")
                    continue
                # Necessary data has been downloaded    
                tide_models_added = True
                added_names.append(s3_name)
        # Write new local bathymetry.tml
        if tide_models_added:
            d = {}
            d["dataset"] = []
            for name in short_name_list:
                d["dataset"].append({"name": name})
            for name in added_names:
                d["dataset"].append({"name": name})
            # Now write the new bathymetry.tml
            with open(os.path.join(self.path, "tide_models.tml"), "w") as tml:
                toml.dump(d, tml)            
            # Read the database again
            self.dataset = []
            self.read()
        # else:
        #     print("No new tide models were added to the local database.")

    def get_dataset(self, name):
        for dataset in self.dataset:
            if dataset.name == name:
                return dataset
        return None

    def dataset_names(self):
        short_name_list = []
        long_name_list = []
        for dataset in self.dataset:
            short_name_list.append(dataset.name)
            long_name_list.append(dataset.long_name)
        return short_name_list, long_name_list


# def dict2yaml(file_name, dct, sort_keys=False):
#     yaml_string = yaml.dump(dct, sort_keys=sort_keys)    
#     file = open(file_name, "w")  
#     file.write(yaml_string)
#     file.close()

# def yaml2dict(file_name):
#     file = open(file_name,"r")
#     dct = yaml.load(file, Loader=yaml.FullLoader)
#     return dct



import pandas as pd
import matplotlib.pyplot as plt

from cht_tide import TideStationsDatabase
from cht_tide import predict

database = TideStationsDatabase(path="c:/work/projects/delftdashboard/delftdashboard_python/data/tide_stations")

dataset = database.dataset["xtide_free"]
dataset.read_data()
gdf = dataset.gdf()

station_name = dataset.station[0]["name"]

prd2 = dataset.predict(name=station_name, start="2023-01-01", end="2023-01-02", dt=300, format="df")

df = dataset.get_components(name=station_name)
# Set times (pandas date range)
times = pd.date_range(start="2023-01-01", end="2023-01-02", freq="10T")
prd = predict(df, times, format="df")
# plot the time series
prd.plot()
# Set the title to station name
plt.title(station_name)
plt.show()



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 10:58:08 2021

@author: Maarten van Ormondt
"""
import os
import xarray as xr

from .model import TideModel

class TideModelFes2014(TideModel):
    """
    """

    def __init__(self, name, path):
        super().__init__()
        
        self.name              = name
        self.path              = path
        # self.local_path        = path
        self.read_metadata()
        self.get_constituents()
    
    def get_constituents(self):
        """
        Get constituents from nc file names in path
        """
        # Loop through nc files in path
        # Get constituents from file names
        filenames = os.listdir(self.path)
        self.constituents = []
        for filename in filenames:
            if filename.endswith(".nc"):
                # Constituent is base name of file
                self.constituents.append(filename.split(".")[0].upper())

    def get_data(self, xl, yl, constituents="all"):
        """
        """
        if constituents == "all":
            constituents = self.constituents

        nconst = len(constituents)

        if xl[0] < 0.0 and xl[1] < 0.0:
            xl = [xl[0] + 360.0, xl[1] + 360.0]

        # Make empty dataset with arrays amplitude and phase, and dimensions lon, lat, and constituent
        ds = xr.Dataset()

        # Get dimensions from first file
        filename = os.path.join(self.path, f"{constituents[0]}.nc")
        ds0 = xr.open_dataset(filename).sel(lon=slice(xl[0], xl[1]), lat=slice(yl[0], yl[1]))
        lon = ds0.lon
        lat = ds0.lat
        ds0.close()
        # Set constituent dimension
        ds["constituent"] = constituents
        # Set lon and lat dimensions
        ds["lon"] = lon
        ds["lat"] = lat
        # Set amplitude and phase arrays
        ds["amplitude"] = xr.DataArray(data=nconst*[len(lat)*[len(lon)*[0.0]]], dims=["constituent", "lat", "lon"])
        ds["phase"] = xr.DataArray(data=nconst*[len(lat)*[len(lon)*[0.0]]], dims=["constituent", "lat", "lon"])

        ds0.close()

        # Loop through constituents
        for constituent in constituents:
            # Get data for constituent
            filename = os.path.join(self.path, f"{constituent}.nc")
            ds0 = xr.open_dataset(filename).sel(lon=slice(xl[0], xl[1]), lat=slice(yl[0], yl[1]))
            # Add data to xarray dataset
            ds["amplitude"].loc[constituent] = ds0["amplitude"].values / 100.0 # Convert cm to m
            ds["phase"].loc[constituent] = ds0["phase"].values
            ds0.close()
        
        return ds




from pathlib import Path

import pandas as pd

from cht_tide.read_bca import SfincsBoundary
from cht_tide.tide_predict import predict


def test_timeseries_from_components():
    # Create dataframe with values
    data = {
        "amplitude": [
            0.39452,
            0.06667,
            0.08968,
            0.01860,
            0.05285,
            0.04300,
            0.01846,
            0.00857,
            0.01022,
            0.00451,
            0.00200,
            0.00101,
            0.00129,
            0.05300,
            0.07800,
        ],
        "phase": [
            14.38,
            40.10,
            352.95,
            36.25,
            217.27,
            234.41,
            220.42,
            233.82,
            348.71,
            347.03,
            180.00,
            89.38,
            202.68,
            50.50,
            176.30,
        ],
    }
    df = pd.DataFrame(
        data,
        index=[
            "M2",
            "S2",
            "N2",
            "K2",
            "K1",
            "O1",
            "P1",
            "Q1",
            "MF",
            "MM",
            "M4",
            "MS4",
            "MN4",
            "SSA",
            "SA",
        ],
    )
    df.index.name = "constituent"

    # Create timeseries
    times = pd.date_range(start="2023-01-01", end="2023-01-02", freq="10T")

    # Predict the tide
    v = predict(df, times)

    assert v.dtype == "float64"
    assert len(times) == len(v)


def test_read_bca_file():
    test_folder = Path().absolute() / "tests"
    bca_file = test_folder.joinpath("sfincs.bca")
    bnd_file = test_folder.joinpath("sfincs.bnd")

    sb = SfincsBoundary()
    sb.read_flow_boundary_points(bnd_file)
    sb.read_astro_boundary_conditions(bca_file)

    assert len(sb.flow_boundary_points) == 28
    assert type(sb.flow_boundary_points[0].astro) == pd.DataFrame


def test_read_bca_and_predict_timeseries():
    test_folder = Path().absolute() / "tests"
    bca_file = test_folder.joinpath("sfincs.bca")
    bnd_file = test_folder.joinpath("sfincs.bnd")

    sb = SfincsBoundary()
    sb.read_flow_boundary_points(bnd_file)
    sb.read_astro_boundary_conditions(bca_file)

    times = pd.date_range(start="2023-01-01", end="2023-01-02", freq="10T")

    # Predict the tide
    v = predict(sb.flow_boundary_points[0].astro, times)

    assert v.dtype == "float64"
    assert len(times) == len(v)



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 10:58:08 2021

@author: Maarten van Ormondt
"""

import os
import numpy as np
import pandas as pd
import toml
import boto3
import xarray as xr
from botocore import UNSIGNED
from botocore.client import Config

import geopandas as gpd
import shapely

from .tide_predict import predict

class TideStationsDataset:
    def __init__(self, name, path):
        self.name = name
        self.long_name = name
        self.path = path
        self.gdf = gpd.GeoDataFrame()
        self.is_read = False
        self.read_metadata()

    def read_metadata(self):
        if not os.path.exists(os.path.join(self.path, "metadata.tml")):
            print("Warning! Tide stations metadata file not found: " + os.path.join(self.path, "metadata.tml"))
            return
        metadata = toml.load(os.path.join(self.path, "metadata.tml"))
        if "longname" in metadata:
            self.long_name = metadata["longname"]
        elif "long_name" in metadata:
            self.long_name = metadata["long_name"]
        self.file = metadata["file"]

    def read_data(self):
        """Reads the netcdf file and store it in self.data"""
        if self.is_read:
            return
        filename = os.path.join(self.path, self.file)
        if not os.path.exists(filename):
            print("Warning! Tide stations dataset file not found: " + filename)
            return
        # Read the netcdf file using xarray
        self.data = xr.open_dataset(filename)
        # Loop through stations
        nr_stations = len(self.data["lon"])
        self.station = []
        for i in range(nr_stations):
            station = {}
            # name
            name = "unknown"
            name_s1 = self.data["stations"][:,i].values
            # convert dtype='|S1' to string
            try:
                name = "".join([x.decode("utf-8") for x in name_s1]).strip()
            except:
                print("Error decoding name")           
            # id
            id_s1 = self.data["idcodes"][:,i].values
            # convert dtype='|S1' to string
            id = "".join([x.decode("utf-8") for x in id_s1]).strip()
            station["name"] = name
            station["id"] = id
            station["lon"] = self.data["lon"][i].values
            station["lat"] = self.data["lat"][i].values
            self.station.append(station)
        # Read the components
        nr_components = np.shape(self.data["components"])[1]
        self.components = []
        for i in range(nr_components):
            components_s1 = self.data["components"][:,i].values
            self.components.append("".join([x.decode("utf-8") for x in components_s1]).strip())

        self.data.close()

        self.is_read = True

    def find_index_by_name(self, name):
        if not self.is_read:
            self.read_data()
        for i, station in enumerate(self.station):
            if station["name"] == name:
                return i
        return None
    
    def find_index_by_id(self, id):
        if not self.is_read:
            self.read_data()
        for i, station in enumerate(self.station):
            if station["id"] == id:
                return i
        return None
    
    def get_components(self, name=None, id=None, index=None, sort=True):
        """Return Pandas dataframe with the tidal components for a station"""
        if name is not None:
            i = self.find_index_by_name(name)
        elif id is not None:
            i = self.find_index_by_id(id)
        else:
            print("Please provide either name or id.")
            return
        if i is None:
            print("Station not found.")
            return
        if not self.is_read:
            self.read_data()
        # Get the amplitudes
        amplitudes = self.data["amplitude"][:,i].values
        # Get the phases
        phases = self.data["phase"][:,i].values
        # Create a dataframe
        df = pd.DataFrame({"constituent": self.components, "amplitude": amplitudes, "phase": phases})
        df = df.set_index("constituent")
        if sort:
            # Sort by amplitude
            df = df.sort_values(by="amplitude", ascending=False)
            # Remove rows with amplitude = 0
            df = df[df.amplitude > 0.0]
        return df

    def predict(self, name=None, id=None, start=None, end=None, t=None, dt=None, offset=0.0, format="tek", filename=None):
        if name is not None:
            components = self.get_components(name=name)
        elif id is not None:
            components = self.get_components(id=id)
        else:
            print("Please provide either name or id.")
            return
        if start is not None and end is not None:
            # if start and end are strings, convert them to datetime objects
            if isinstance(start, str):
                start = pd.to_datetime(start)
            if isinstance(end, str):
                end = pd.to_datetime(end)
            if dt is None:
                dt = 600.0 # seconds
            times = pd.date_range(start=start, end=end, freq=f"{dt}s")
        elif t is not None:
            times = t
        else:
            print("Please provide either t0 and t1, or t.")
            return
        prd = predict(components, times, format="df")
        prd = prd + offset
        if filename is not None:
            if format == "tek":
                # Write to a .tek file
                df2tekaltimeseries(prd, filename)
            elif format == "csv":
                # Write to a .csv file, skipping the header
                prd.to_csv(filename, header=False)
        return prd
        
    def get_gdf(self):
        if not self.is_read:
            self.read_data()
        if len(self.gdf) == 0:    
            gdf_list = []
            # Loop through points
            for station in self.station:
                point = shapely.geometry.Point(station["lon"], station["lat"])
                d = {"id": station["id"], "name": station["name"], "geometry": point}
                gdf_list.append(d)
            self.gdf = gpd.GeoDataFrame(gdf_list, crs=4326)
        return self.gdf
    
    def station_names(self):
        """Return lists of station names and ids"""
        name_list = []
        id_list = []
        # Loop through the keys of the dictionary
        for station in self.station:
            name_list.append(station["name"])
            id_list.append(station["id"])
        return name_list, id_list

class TideStationsDatabase:
    """
    The main Tide Stations Database class
    
    :param pth: Path name where bathymetry tiles will be cached.
    :type pth: string            
    """
    
    def __init__(self, path=None,
                 s3_bucket=None,
                 s3_key=None,
                 s3_region=None):
        self.path    = path
        self.dataset = {}
        self.s3_client = None
        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.s3_region = s3_region
        self.read()
    
    def read(self):
        """
        Reads meta-data of all datasets in the database. 
        """
        if self.path is None:
            print("Path to tide stations database not set !")
            return
        
        # Check if the path exists. If not, create it.
        if not os.path.exists(self.path):
            os.makedirs(self.path)

        # Read in database
        tml_file = os.path.join(self.path, "tide_stations.tml")
        if not os.path.exists(tml_file):
            print("Warning! Tide stations database file not found: " + tml_file)
            return

        datasets = toml.load(tml_file)

        for d in datasets["dataset"]:

            name = d["name"]

            if "path" in d:
                path = d["path"]
            else:
                path = os.path.join(self.path, name)

            # # Read the meta data for this dataset
            # fname = os.path.join(path, "metadata.tml")

            # if os.path.exists(fname):
            #     metadata = toml.load(fname)
            #     dataset_format = metadata["format"]
            # else:
            #     print("Could not find metadata file for dataset " + name + " ! Skipping dataset.")
            #     continue
             
            self.dataset[name] = TideStationsDataset(name, path)

    def check_online_database(self):
        if self.s3_client is None:
            self.s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
        if self.s3_bucket is None:
            return
        # First download a copy of bathymetry.tml and call it bathymetry_s3.tml
        key = f"{self.s3_key}/tide_stations.tml"
        filename = os.path.join(self.path, "tide_stations_s3.tml")
        print("Updating tide stations database ...")
        try:
            self.s3_client.download_file(Bucket=self.s3_bucket,     # assign bucket name
                                         Key=key,           # key is the file name
                                         Filename=filename) # storage file path
        except:
            # Download failed
            print(f"Failed to download {key} from {self.s3_bucket}. Database will not be updated.")
            return

        # Read bathymetry_s3.tml
        short_name_list, long_name_list = self.dataset_names()
        datasets_s3 = toml.load(filename)
        tide_stations_added = False
        added_names = []
        # Loop through s3 datasets, and check whether they exist in the local database.
        # If so, check if the metadata also exists. If not, make local folder and download the metadata.
        # Additionally, check if available_tiles.nc in s3 and not in local database, download it.
        for d in datasets_s3["dataset"]:
            # Get list of existing datasets
            s3_name = d["name"]
            if s3_name not in short_name_list:
                # Dataset not in local database
                print(f"Adding tide stations {s3_name} to local database ...")
                # Create folder and download metadata
                path = os.path.join(self.path, s3_name)
                os.makedirs(path, exist_ok=True)
                key = f"{self.s3_key}/{s3_name}/metadata.tml"
                filename = os.path.join(path, "metadata.tml")
                # Download metadata
                try:
                    self.s3_client.download_file(Bucket=self.s3_bucket, # assign bucket name
                                                Key=key,               # key is the file name
                                                Filename=filename)     # storage file path
                except Exception as e:
                    print(e)
                    print(f"Failed to download {key}. Skipping tide stations dataset.")
                    continue
                # Necessary data has been downloaded    
                tide_stations_added = True
                added_names.append(s3_name)
        # Write new local bathymetry.tml
        if tide_stations_added:
            d = {}
            d["dataset"] = []
            for name in short_name_list:
                d["dataset"].append({"name": name})
            for name in added_names:
                d["dataset"].append({"name": name})
            # Now write the new bathymetry.tml
            with open(os.path.join(self.path, "tide_stations.tml"), "w") as tml:
                toml.dump(d, tml)            
            # Read the database again
            self.dataset = {}
            self.read()
        # else:
        #     print("No new tide models were added to the local database.")

    # def get_dataset(self, name):
    #     for dataset in self.dataset:
    #         if dataset.name == name:
    #             return dataset
    #     return None

    def dataset_names(self):
        short_name_list = []
        long_name_list = []
        # self.dataset is a dictionary
        # Loop through the keys of the dictionary
        for key in self.dataset.keys():
            short_name_list.append(key)
            long_name_list.append(self.dataset[key].long_name)
        # for dataset in self.dataset:
        #     short_name_list.append(dataset.name)
        #     long_name_list.append(dataset.long_name)
        return short_name_list, long_name_list


# def dict2yaml(file_name, dct, sort_keys=False):
#     yaml_string = yaml.dump(dct, sort_keys=sort_keys)    
#     file = open(file_name, "w")  
#     file.write(yaml_string)
#     file.close()

# def yaml2dict(file_name):
#     file = open(file_name,"r")
#     dct = yaml.load(file, Loader=yaml.FullLoader)
#     return dct

def df2tekaltimeseries(df, filename):
    """Write a Pandas dataframe to a .tek file"""
    nt = len(df)
    # Convert index of dataframe to string with format "YYYYMMDD HHMMSS"
    indexstr = df.index.strftime("%Y%m%d %H%M%S")
    with open(filename, "w") as f:
        f.write("* column 1 : Date\n")
        f.write("* column 2 : Time\n")
        f.write("* column 3 : WL\n")
        f.write("BL01\n")
        f.write(f"{nt} 4\n")
        j = 0
        for i, row in df.iterrows():
            # date, time and first column of data
            tstr = indexstr[j]
            j += 1
            vstr = f"{row[0]:7.3f}"
            f.write(f"{tstr} {vstr}\n")


# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 10:58:08 2021

@author: Maarten van Ormondt
"""
import os
import toml
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
from pyproj import CRS
import boto3
from botocore import UNSIGNED
from botocore.client import Config
from  cht_utils import fileops as fo

class TideModel:
    """
    Tide model 
    """
    def __init__(self):        
        self.database          = None
        self.name              = ""
        self.long_name         = ""
        self.path              = ""
        self.main_constituents = ["M2", "S2", "N2", "K2", "K1", "O1", "P1", "Q1"]
        self.files             = []

    def read_metadata(self):
        # Read metadata file
        tml_file = os.path.join(self.path, "metadata.tml")
        tml = toml.load(tml_file)
        for key in tml:
            setattr(self, key, tml[key])
        # Long name for backwards compatibility
        if "longname" in tml:
            self.long_name = tml["longname"]
        # Make sure there is always a long_name
        if self.long_name == "":
            self.long_name = self.name        

        self.crs = CRS(4326)

    def download(self):
        if self.s3_bucket is None:
            return
        # Check if download is needed
        for file in self.files:
            if not os.path.exists(os.path.join(self.path, file)):
                s3_client = boto3.client('s3', config=Config(signature_version=UNSIGNED))
                break
        # Get all files defined in the toml file
        for file in self.files:
            if not os.path.exists(os.path.join(self.path, file)):
                print(f"Downloading {file} from tide model {self.name} ...")
                s3_client.download_file(self.s3_bucket, f"{self.s3_key}/{file}", os.path.join(self.path, file))
            
    def get_data_on_points(self, gdf=None, x=None, y=None, crs=None, format="gdf", constituents="all"):
        """
        x can be a list of x coordinates, or and array of x coordinates
        y can be a list of y coordinates, or and array of y coordinates
        """

        # Download files if needed
        self.download()

        # Return pandas dataframe with constituents as rows and amplitudes and phases as columns

        if constituents == "all":
            constituents = self.constituents
        elif constituents == "main":
            constituents = self.main_constituents

        if gdf is not None:
            # Transform gdf to lon, lat (that's what the tide model uses) 
            gdf4326 = gdf.to_crs("EPSG:4326")
            # Get extent of gdf
            xl = [gdf4326.geometry.x.min(), gdf4326.geometry.x.max()]
            yl = [gdf4326.geometry.y.min(), gdf4326.geometry.y.max()]
            # Add a little buffer
            xl[0] -= 0.25
            xl[1] += 0.25
            yl[0] -= 0.25
            yl[1] += 0.25 
        else:
            # Make gdf from x and y
            gdf = pd.DataFrame()
            gdf["geometry"] = [Point(x, y) for x, y in zip(x, y)]
            gdf = gpd.GeoDataFrame(gdf, crs=crs)     

        # Get the data in the extent
        ds = self.get_data(xl, yl, constituents=constituents)

        if format == "gdf" or format == "geodataframe":
            if "astro" not in gdf.columns:
                gdf["astro"] = None
            # Create geodataframe with points
            # Loop over points
            for i, row in gdf.to_crs("EPSG:4326").iterrows():
                x = np.mod(row.geometry.x, 360.0)
                y = row.geometry.y
                # First convert tidal data to vector
                ds["tvu"] = ds.amplitude * np.cos(ds.phase*np.pi/180.0)
                ds["tvv"] = ds.amplitude * np.sin(ds.phase*np.pi/180.0)
                # Interpolate
                dsp = ds.interp(lon=x, lat=y)
                df = pd.DataFrame()
                df["constituent"] = constituents
                # Now convert back to amplitude and phase
                dsp["amplitude"] = np.sqrt(dsp.tvu**2 + dsp.tvv**2)
                dsp["phase"] = np.mod(np.arctan2(dsp.tvv, dsp.tvu) * 180.0 / np.pi, 360.0)
                df["amplitude"] = dsp.amplitude.values
                df["phase"] = dsp.phase.values
                df = df.set_index("constituent")
                gdf.at[i, "astro"] = df
            return gdf
        elif format == "dataframe" or format == "df" or format == "pandas":
            # Return list with dataframes
            lst = []
            for i, row in gdf.to_crs("EPSG:4326").iterrows():
                dsp = ds.interp(lon=np.array(row.geometry.x), lat=np.array(row.geometry.y))
                df = pd.DataFrame()
                df["constituent"] = constituents
                df["amplitude"] = dsp.amplitude.values
                df["phase"] = dsp.phase.values
                df.set_index("constituent")
                lst.append(df)
            return lst



#   Copyright notice
#   --------------------------------------------------------------------
#   Copyright (C) 2020 Deltares
#       Freek Scheel
#
#       freek.scheel@deltares.nl
#
#       P.O. Box 177
#       2600 MH Delft
#       The Netherlands
#
#   This library is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This library is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this library.  If not, see <http://www.gnu.org/licenses/>.
#   --------------------------------------------------------------------

import numpy as np

d2r, r2d = np.pi / 180.0, 180.0 / np.pi

# The following functions take a dictionary of astronomical values (in degrees)
# and return dimensionless scale factors for constituent amplitudes.


def f_unity(a):
    return 1.0


# Schureman equations 73, 65
def f_Mm(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    mean = (2 / 3.0 - np.sin(omega) ** 2) * (1 - 3 / 2.0 * np.sin(i) ** 2)
    return (2 / 3.0 - np.sin(I) ** 2) / mean


# Schureman equations 74, 66
def f_Mf(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    mean = np.sin(omega) ** 2 * np.cos(0.5 * i) ** 4
    return np.sin(I) ** 2 / mean


# Schureman equations 75, 67
def f_O1(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    mean = np.sin(omega) * np.cos(0.5 * omega) ** 2 * np.cos(0.5 * i) ** 4
    return (np.sin(I) * np.cos(0.5 * I) ** 2) / mean


# Schureman equations 76, 68
def f_J1(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    mean = np.sin(2 * omega) * (1 - 3 / 2.0 * np.sin(i) ** 2)
    return np.sin(2 * I) / mean


# Schureman equations 77, 69
def f_OO1(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    mean = np.sin(omega) * np.sin(0.5 * omega) ** 2 * np.cos(0.5 * i) ** 4
    return np.sin(I) * np.sin(0.5 * I) ** 2 / mean


# Schureman equations 78, 70
def f_M2(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    mean = np.cos(0.5 * omega) ** 4 * np.cos(0.5 * i) ** 4
    return np.cos(0.5 * I) ** 4 / mean


# Schureman equations 227, 226, 68
# Should probably eventually include the derivations of the magic numbers (0.5023 etc).
def f_K1(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    nu = d2r * a["nu"].value
    sin2Icosnu_mean = np.sin(2 * omega) * (1 - 3 / 2.0 * np.sin(i) ** 2)
    mean = 0.5023 * sin2Icosnu_mean + 0.1681
    return (
        0.2523 * np.sin(2 * I) ** 2 + 0.1689 * np.sin(2 * I) * np.cos(nu) + 0.0283
    ) ** (0.5) / mean


# Schureman equations 215, 213, 204
# It can be (and has been) confirmed that the exponent for R_a reads 1/2 via Schureman Table 7
def f_L2(a):
    P = d2r * a["P"].value
    I = d2r * a["I"].value
    R_a_inv = (
        1 - 12 * np.tan(0.5 * I) ** 2 * np.cos(2 * P) + 36 * np.tan(0.5 * I) ** 4
    ) ** (0.5)
    return f_M2(a) * R_a_inv


# Schureman equations 235, 234, 71
# Again, magic numbers
def f_K2(a):
    omega = d2r * a["omega"].value
    i = d2r * a["i"].value
    I = d2r * a["I"].value
    nu = d2r * a["nu"].value
    sinsqIcos2nu_mean = np.sin(omega) ** 2 * (1 - 3 / 2.0 * np.sin(i) ** 2)
    mean = 0.5023 * sinsqIcos2nu_mean + 0.0365
    return (
        0.2533 * np.sin(I) ** 4 + 0.0367 * np.sin(I) ** 2 * np.cos(2 * nu) + 0.0013
    ) ** (0.5) / mean


# Schureman equations 206, 207, 195
def f_M1(a):
    P = d2r * a["P"].value
    I = d2r * a["I"].value
    Q_a_inv = (
        0.25
        + 1.5 * np.cos(I) * np.cos(2 * P) * np.cos(0.5 * I) ** (-0.5)
        + 2.25 * np.cos(I) ** 2 * np.cos(0.5 * I) ** (-4)
    ) ** (0.5)
    return f_O1(a) * Q_a_inv


# See e.g. Schureman equation 149
def f_Modd(a, n):
    return f_M2(a) ** (n / 2.0)


# Node factors u, see Table 2 of Schureman.


def u_zero(a):
    return 0.0


def u_Mf(a):
    return -2.0 * a["xi"].value


def u_O1(a):
    return 2.0 * a["xi"].value - a["nu"].value


def u_J1(a):
    return -a["nu"].value


def u_OO1(a):
    return -2.0 * a["xi"].value - a["nu"].value


def u_M2(a):
    return 2.0 * a["xi"].value - 2.0 * a["nu"].value


def u_K1(a):
    return -a["nup"].value


# Schureman 214
def u_L2(a):
    I = d2r * a["I"].value
    P = d2r * a["P"].value
    R = r2d * np.arctan(
        np.sin(2 * P) / (1 / 6.0 * np.tan(0.5 * I) ** (-2) - np.cos(2 * P))
    )
    return 2.0 * a["xi"].value - 2.0 * a["nu"].value - R


def u_K2(a):
    return -2.0 * a["nupp"].value


# Schureman 202
def u_M1(a):
    I = d2r * a["I"].value
    P = d2r * a["P"].value
    Q = r2d * np.arctan((5 * np.cos(I) - 1) / (7 * np.cos(I) + 1) * np.tan(P))
    return a["xi"].value - a["nu"].value + Q


def u_Modd(a, n):
    return n / 2.0 * u_M2(a)



# -*- coding: utf-8 -*-
"""
Created on Sun Apr 25 10:58:08 2021

@author: ormondt
"""
__version__ = "0.1.0"

from .model import TideModel
from .database import TideModelDatabase
from .tide_stations import TideStationsDatabase
from .tide_predict import predict



import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import Point

from cht_tide.fes2014 import TideModelFes2014
from cht_tide.tide_predict import predict

name = "fes2014"
path = r"c:\work\delftdashboard\data\tidemodels\fes2014"
mdl = TideModelFes2014(name, path)
ds=mdl.get_data([350.0, 351.0], [50.0, 51.0], constituents=["m2"])

x = [350.0, 351.0, 352.0]
y = [50.0, 51.0, 52.0]

# Make GeoDataFrame with points
gdf = GeoDataFrame(
    {"geometry": [Point(x, y) for x, y in zip(x, y)]},
    crs="EPSG:4326",
)

#ds=mdl.get_data_on_points(gdf=gdf, constituents=["m2"])
gdf = mdl.get_data_on_points(gdf=gdf, constituents="all")

# Create timeseries
times = pd.date_range(start="2023-01-01", end="2023-01-02", freq="10T")

    # Predict the tide

for i, point in gdf.iterrows():
    v = predict(point["astro"], times)

assert v.dtype == "float64"
assert len(times) == len(v)
